/*
 *  Copyright (c) 2007 Michael Heimpold <m.heimpold@web.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2
 *  as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program (see the file COPYING included with this
 *  distribution); if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* $Id: main.c 7 2007-07-10 10:28:46Z mhei $ */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "linux_avm_cpmac.h"
#include "adm6996.h"
#include "config.h"

#define MAX_SWITCH_PORTS 4

#ifdef DEBUG
#define DEBUGMSG(a...) fprintf(stderr, "[debug] " a);
#else
#define DEBUGMSG(a...)
#endif

char *cpmac_modes[CPMAC_MODE_MAX_NO] =
  {"normal", "ata", "split", "split_ata", "all_ports", "special"};

char *cpmac_adm_phy_power_modes[] =
  {"off", "save", "on"};

enum media_type {
  MEDIA_AUTO  = 0,
  MEDIA_10HD  = 1,
  MEDIA_10FD  = 2,
  MEDIA_100HD = 3,
  MEDIA_100FD = 4,
  MEDIA_COUNT = 5,
};
char *media_str[] =
  {"auto", "10baseT-HD", "10baseT-FD", "100baseTx-HD", "100baseTx-FD"};

enum flow_control {
  FC_DONTTOUCH,
  FC_ENABLE,
  FC_DISABLE,
  FC_MAX,
};
char *flow_control_str[] =
  {"donttouch", "enable", "disable"};

char devname[AVM_CPMAC_MAX_DEVICE_NAME_LENGTH];
int sockfd;

void initsockfd(void) {
  if ((sockfd = socket(PF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("socket");
    exit(EXIT_FAILURE);
  }
}

void prepare_ifreq(struct ifreq *ifr, void *data) {
  memset(ifr, 0, sizeof(*ifr));
  strncpy(ifr->ifr_name, devname, IFNAMSIZ);
  ifr->ifr_data = data;
}


int count_ports(unsigned int *internal, unsigned int *external) {
  struct ifreq ifr;
  struct avm_switch_info data;
  int rv;

  memset(&data, 0, sizeof(data));

  prepare_ifreq(&ifr, (void *)&data);

  rv = ioctl(sockfd, AVM_CPMAC_IOCTL_INFO, &ifr);
  
  if (internal)
    *internal = data.internal_ports;
  if (external)
    *external = data.external_ports;

  return rv;
}

int do_info(void) {
  unsigned int internal, external;

  if (count_ports(&internal, &external) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  printf("Internal ports: %u\nExternal ports: %u\n",
         internal, external);
  return EXIT_SUCCESS;
}


int do_get_switch_mode(void) {
  struct avm_cpmac_ioctl_struct data;
  struct ifreq ifr;
  int ret;
  unsigned int i;

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_GET_SWITCH_MODE;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    DEBUGMSG("AVM_CPMAC_IOCTL_GENERIC failed with ret=%d, trying AVM_CPMAC_IOCTL_GET_CONFIG_N\n", ret);
    prepare_ifreq(&ifr, (void *)&data.u.switch_mode);
    ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GET_CONFIG_N, &ifr);
    if (ret != 0) {
      switch (ret) {
        case -EIO:
          fprintf(stderr, "Driver reported: no VLAN possible.\n");
          break;
        default:
          perror("ioctl");
      }
      return EXIT_FAILURE;
    }
  }
  printf("Current mode: %s\n", cpmac_modes[data.u.switch_mode.cpmac_mode - 1]);
  printf("VLAN IDs passed through WAN:");
  if (data.u.switch_mode.wan_vlan_number == 0) {
    printf(" (none)\n");
  } else {
    for (i = 0; i < data.u.switch_mode.wan_vlan_number; ++i) {
      printf(" %#x", data.u.switch_mode.wan_vlan_id[i]);
    }
    printf("\n");
  }
  return EXIT_SUCCESS;
}


int do_set_switch_mode(struct avm_cpmac_config_struct *mode) {
  struct ifreq ifr;
  struct avm_cpmac_ioctl_struct data;
  int ret;

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_SET_SWITCH_MODE;
  memcpy(&data.u.switch_mode, mode, sizeof(data.u.switch_mode));

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    DEBUGMSG("AVM_CPMAC_IOCTL_GENERIC failed with ret=%d, trying AVM_CPMAC_IOCTL_SET_CONFIG_N\n", ret);
    prepare_ifreq(&ifr, (void *)&data.u.switch_mode);
    ret = ioctl(sockfd, AVM_CPMAC_IOCTL_SET_CONFIG_N, &ifr);
    if (ret != 0) {
      switch (ret) {
        case -EIO:
          fprintf(stderr, "Driver reported: no VLAN possible.\n");
          break;
        default:
          perror("ioctl");
      }
      return EXIT_FAILURE;
    }
  }
  return EXIT_SUCCESS;
}


int do_ioctl_generic(unsigned int type, void *data, size_t n) {
  struct ifreq ifr;
  struct avm_cpmac_ioctl_struct aci;
  int ret;

  memset(&aci, 0, sizeof(aci));
  aci.type = type;
  memcpy(&aci.u, data, n < sizeof(aci.u) ? n : sizeof(aci.u));

  prepare_ifreq(&ifr, (void *)&aci);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    switch (ret) {
      case -EIO:
        fprintf(stderr, "Driver reported: no VLAN possible.\n");
        break;
      default:
        perror("ioctl");
    }
  } else {
    memcpy(data, &aci.u, n);
  }
  return ret;
}


int do_set_switch_mode_special(cpmac_switch_configuration_t *switch_config) {
  int ret;

  ret = do_ioctl_generic(AVM_CPMAC_IOCTL_CONFIG_SET_SWITCH_MODE_SPECIAL,
                         (void *)switch_config,
                         sizeof(*switch_config));

  return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}


int do_get_switch_mode_current(void) {
  cpmac_switch_configuration_t switch_config;
  int ret, i;

  ret = do_ioctl_generic(AVM_CPMAC_IOCTL_CONFIG_GET_SWITCH_MODE_CURRENT,
                         (void *)&switch_config,
                         sizeof(switch_config));

  if (ret != 0)
    return EXIT_FAILURE;

  printf("Devices: %u\n", switch_config.number_of_devices);
  printf("WAN is port: ");
  if (switch_config.wanport != 0xff) {
    printf("%u", switch_config.wanport);
  } else {
    printf("(none)");
  }
  printf("\n");

  for (i = 0; i < switch_config.number_of_devices; ++i)
    printf("Device %u: name=%s, portmask=0x%02x\n",
           i + 1,
           switch_config.device[i].name,
           switch_config.device[i].target_mask);
  return EXIT_SUCCESS;
}


int do_get_power_mode(void) {
  struct avm_cpmac_ioctl_struct data;
  struct ifreq ifr;
  int ret, i;
  int c = sizeof(data.u.phy_power_setup.mode) / sizeof(data.u.phy_power_setup.mode[0]);
  unsigned int ports;
  
  if (count_ports(NULL, &ports) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }

  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_GET_PHY_POWER;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    DEBUGMSG("AVM_CPMAC_IOCTL_GENERIC failed with ret=%d, trying AVM_CPMAC_IOCTL_GET_CONFIG_N\n", ret);
    prepare_ifreq(&ifr, (void *)&data.u.phy_power_setup);
    ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GET_CONFIG_N, &ifr);
    if (ret != 0) {
      perror("ioctl");
      return EXIT_FAILURE;
    }
  }

  DEBUGMSG("Time On:  %u\n", data.u.phy_power_setup.time_on);
  DEBUGMSG("Time Off: %u\n", data.u.phy_power_setup.time_off);
  for (i = 0; i < c && i < ports; ++i)
    printf("Port %u: %s\n",
           i + 1,
	   cpmac_adm_phy_power_modes[data.u.phy_power_setup.mode[i]]);
  return EXIT_SUCCESS;
}


int do_set_power_mode(unsigned int port, unsigned int new_pm) {
  struct avm_cpmac_ioctl_struct data;
  struct ifreq ifr;
  int ret, i;
  int c = sizeof(data.u.phy_power_setup.mode) / sizeof(data.u.phy_power_setup.mode[0]);
  unsigned int ports;
  int use_fallback = 0;
  
  if (count_ports(NULL, &ports) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  
  if (port < 1 || port > ports) {
    fprintf(stderr, "Unknown port. This box has (only) %u external port%s.\n", ports,
            ports > 1 ? "s" : "");
    return EXIT_FAILURE;
  }
  
  /* get current config */
  memset(&data, 0, sizeof(data));
  data.type = AVM_CPMAC_IOCTL_CONFIG_GET_PHY_POWER;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    DEBUGMSG("AVM_CPMAC_IOCTL_GENERIC failed with ret=%d, trying AVM_CPMAC_IOCTL_GET_CONFIG_N\n", ret);
    prepare_ifreq(&ifr, (void *)&data.u.phy_power_setup);
    ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GET_CONFIG_N, &ifr);
    if (ret != 0) {
      perror("ioctl");
      return EXIT_FAILURE;
    }
    use_fallback = 1;
  }

  /* set new config */
  data.type = AVM_CPMAC_IOCTL_CONFIG_SET_PHY_POWER;
  data.u.phy_power_setup.mode[port - 1] = new_pm;
  
  if (use_fallback)
    ret = ioctl(sockfd, AVM_CPMAC_IOCTL_SET_CONFIG_N, &ifr);
  else
    ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr);
  if (ret != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}


int switchreg(int set, unsigned int reg, unsigned int *val) {
  struct avm_cpmac_ioctl_struct data;
  struct ifreq ifr;
  int ret;

  memset(&data, 0, sizeof(data));
  data.type = set ? AVM_CPMAC_IOCTL_CONFIG_SET_SWITCH_REGISTER : AVM_CPMAC_IOCTL_CONFIG_GET_SWITCH_REGISTER;
  data.u.switch_register.reg = reg;
  if (set && val)
    data.u.switch_register.value = *val;

  prepare_ifreq(&ifr, (void *)&data);

  if ((ret = ioctl(sockfd, AVM_CPMAC_IOCTL_GENERIC, &ifr)) != 0) {
    DEBUGMSG("AVM_CPMAC_IOCTL_GENERIC failed with ret=%d, trying AVM_CPMAC_IOCTL_{G,S}ET_CONFIG_N\n", ret);
    prepare_ifreq(&ifr, (void *)&data.u);
    ret = ioctl(sockfd, set ? AVM_CPMAC_IOCTL_SET_CONFIG_N : AVM_CPMAC_IOCTL_GET_CONFIG_N, &ifr);
  }

  if (ret == 0 && !set && val)
    *val = data.u.switch_register.value;

  return ret;
}


int do_get_port_media(unsigned int port) {
  unsigned int val1, shift;
  unsigned int reg2, val2;
  
  switch (port) {
    case 1: shift = ADM_STATUS_PORT0_SHIFT; reg2 = REG_ADM_LC_PORT0_CONF; break;
    case 2: shift = ADM_STATUS_PORT1_SHIFT; reg2 = REG_ADM_LC_PORT1_CONF; break;
    case 3: shift = ADM_STATUS_PORT2_SHIFT; reg2 = REG_ADM_LC_PORT2_CONF; break;
    case 4: shift = ADM_STATUS_PORT3_SHIFT; reg2 = REG_ADM_LC_PORT3_CONF; break;
    default:
	    fprintf(stderr, "Unknown port.\n");
	    return EXIT_FAILURE;
  }

  if (switchreg(0, REG_ADM_LC_STATUS0, &val1) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  if (switchreg(0, REG_ADM_LC_STATUS0 + 1, &val2) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  val1 = (val2 << 16 | val1) >> shift;
  if (switchreg(0, reg2, &val2) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  if (val1 & ADM_STATUS_LINKUP)
    printf("Port %u:%s 10%sbaseT%s-%sD%s, %s\n", port,
           val2 & ADM_CONF_AUTO_NEGOTIATION ? " negotiated" : " no autonegotiation,",
           val1 & ADM_STATUS_SPEED ? "0" : "",
           val1 & ADM_STATUS_SPEED ? "x" : "",
	   val1 & ADM_STATUS_DUPLEX ? "F" : "H",
	   val1 & ADM_STATUS_FLOWCONTROL ? " flow-control" : "",
	   val1 & ADM_STATUS_LINKUP ? "link ok" : "no link");
  else
    printf("Port %u: no link\n", port);

  return EXIT_SUCCESS;
}


int do_set_port_media(unsigned int port, enum media_type media, enum flow_control fc) {
  unsigned int val;
  unsigned int reg;
  
  switch (port) {
    case 1: reg = REG_ADM_LC_PORT0_CONF; break;
    case 2: reg = REG_ADM_LC_PORT1_CONF; break;
    case 3: reg = REG_ADM_LC_PORT2_CONF; break;
    case 4: reg = REG_ADM_LC_PORT3_CONF; break;
    default:
	    fprintf(stderr, "Unknown port.\n");
	    return EXIT_FAILURE;
  }
  if (switchreg(0, reg, &val) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  switch (fc) {
    case FC_ENABLE:  val |= ADM_CONF_FLOW_CONTROL; break;
    case FC_DISABLE: val &= ~ADM_CONF_FLOW_CONTROL; break;
  }
  val &= ~ADM_CONF_AUTO_NEGOTIATION;
  switch (media) {
      case MEDIA_10HD:
        val &= ~(ADM_CONF_DUPLEX_FULL | ADM_CONF_SPEED_100);
	break;
      case MEDIA_10FD:
        val |= ADM_CONF_DUPLEX_FULL;
        val &= ~ADM_CONF_SPEED_100;
	break;
      case MEDIA_100HD:
	val &= ~ADM_CONF_DUPLEX_FULL;
        val |= ADM_CONF_SPEED_100;
	break;
      case MEDIA_100FD:
        val |= ADM_CONF_DUPLEX_FULL | ADM_CONF_SPEED_100;
        break;
      default:
        val |= ADM_CONF_DUPLEX_FULL | ADM_CONF_SPEED_100 | ADM_CONF_AUTO_NEGOTIATION;
  }
  if (switchreg(1, reg, &val) != 0) {
    perror("ioctl");
    return EXIT_FAILURE;
  }
  return EXIT_SUCCESS;
}


void version(void) {
  fprintf(stderr,
    "cpmaccfg (%s)\n"
    "Copyright (c) 2007 Michael Heimpold <%s>.\n"
    "\n"
    "This is free software; see the source for copying conditions.  There is NO\n"
    "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
    PACKAGE_VERSION, PACKAGE_BUGREPORT);
  exit(EXIT_SUCCESS);
}


void usage(int argc, char *argv[]) {
  fprintf(stderr, "Usage: %s [OPTIONS]... COMMAND [ARGS]\n"
  "\n"
  "Show/change VLAN configuration of avm_cpmac driver.\n"
  "\n"
  "  Options:\n"
  "    -i,--iface            interface to operate on (default: eth0)\n"
  "    -p,--pass-to-wan ID   VLAN ID passed through WAN port (only useful for\n"
  "                          a set_switch_mode command resulting in a WAN port;\n"
  "                          repeat this option to pass multiple IDs)\n"
  "    -w,--wan-port NUM     set WAN port number for set_switch_mode_special\n"
  "                          1 is first switch port,\n"
  "                          255 means no port is WAN (default)\n"
  "\n"
  "    -h,--help             display this help and exit\n"
  "       --version          output version information and exit\n"
  "\n"
  "  Commands:\n"
  "  info                    print count of internal and external ports\n"
  "\n"
  "  get_switch_mode|gsm     print current switch mode and VLAN IDs passed to WAN\n"
  "\n"
  "  set_switch_mode|ssm NEWMODE     set switch mode to NEWMODE\n"
  "    Implemented modes are (as stated in linux_avm_cpmac.h):\n"
  "      mode        7170             VDSL             ProfiVoIP\n"
  "      normal      eth0             eth0, wan        eth0\n"
  "      ata         eth0, wan        eth0, wan        eth0, wan\n"
  "      split       eth0-eth3        eth0-eth3, wan   eth0-eth4\n"
  "      split_ata   eth0-eth2, wan   eth0-eth2, wan   eth0-eth3, wan\n"
  "      all_ports   eth0             eth0             eth0\n"
  "      special     ???              ???              ???\n"
  "    By default, mode special is a placeholder which have to be filled by issuing\n"
  "    a set_switch_mode_special command first which a custom configuration.\n"
  "    In configurations that provide a WAN port, VLANs can be routed through them,\n"
  "    see option --pass-to-wan.\n"
  "\n"
  "  set_switch_mode_special|ssms DEVICENAME PORTMASK [DEVICENAME PORTMASK]...\n"
  "    Set a special switch mode configuration (see also option --wan):\n"
  "      DEVICENAME is the name of the (new) network device\n"
  "      PORTMASK specifies which switch port(s) the device is connected to\n"
  "      port1=0x01, port2=0x02, port3=0x04, port4=0x08, cpu port=0x20\n"
  "    Example: mode 'ata' on a 7170 is \"... -w 1 eth0 0x2e wan 0x21\"\n"
  "\n"
  "  get_switch_mode_current|gsmc    print current switch configuration\n"
  "\n"
  "  get_power_mode|gpm              print current power mode configuration\n"
  "\n"
  "  set_power_mode|spm PORTNUMBER POWERMODE\n"
  "    Set power mode for given port number where POWERMODE is one of:\n"
  "    'off' - disable port, 'save' - auto-detect, 'on' - always on\n"
  "\n"
  "  get_port_media|gpme [PORTNUMBER]\n"
  "    Prints current power media configuration\n"
  "\n"
  "  set_port_media|spme PORTNUMBER MEDIA [FLOWCONTROL]\n"
  "    Force MEDIA type for port PORTNUMBER. Media is one of the following strings:\n"
  "    auto, 10baseT-HD, 10baseT-FD, 100baseTx-HD, 100baseTx-FD\n"
  "    FLOW-CONTROL: donttouch (default), enable, disable\n"
  "\n"
  "Report bugs to <%s>.\n",
  argc > 0 ? argv[0] : PACKAGE, PACKAGE_BUGREPORT);
  exit(EXIT_FAILURE);
}


int main(int argc, char *argv[]) {
  int c;
  unsigned int tmp;
  struct avm_cpmac_config_struct param_ssm;
  cpmac_switch_configuration_t param_ssms;
  unsigned int max_wan_vlan_ids = sizeof(param_ssm.wan_vlan_id) / sizeof(param_ssm.wan_vlan_id[0]);

  memset(devname, 0, sizeof(devname));
  strncpy(devname, "eth0", sizeof(devname) - 1);
  memset(&param_ssm, 0, sizeof(param_ssm));
  memset(&param_ssms, 0, sizeof(param_ssms));
  param_ssms.wanport = 0xff;

  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"help", 0, NULL, 'h'},
      {"iface", 1, NULL, 'i'},
      {"pass-to-wan", 1, NULL, 'p'},
      {"version", 0, NULL, 'v'},
      {"wan-port", 1, NULL, 'w'},
      {0, 0, 0, 0}
    };
    c = getopt_long(argc, argv, "hi:p:w:", long_options, &option_index);
    if (c == -1)
      break;
    switch (c) {
      case 'h':
        usage(argc, argv);
      case 'i':
        strncpy(devname, optarg, sizeof(devname) - 1);
        break;
      case 'p':
        if (param_ssm.wan_vlan_number >= max_wan_vlan_ids) {
          fprintf(stderr, "Warning: Maximum of %u VLAN IDs reached, ignoring VLAN ID %s.\n",
                  max_wan_vlan_ids, optarg);
          continue;
        }
        if (sscanf(optarg, "%u", &param_ssm.wan_vlan_id[param_ssm.wan_vlan_number]) != 1) {
          fprintf(stderr, "Error: Can't parse VLAN ID '%s'.\n", optarg);
          exit(EXIT_FAILURE);
        }
        ++param_ssm.wan_vlan_number;
        break;
      case 'v':
        version();
      case 'w':
        if (sscanf(optarg, "%u", &tmp) != 1) {
          fprintf(stderr, "Error: Can't parse WAN port number '%s'.\n", optarg);
          exit(EXIT_FAILURE);
        }
        if ((tmp >= 1 && tmp <= MAX_SWITCH_PORTS) || tmp == 255) {
          param_ssms.wanport = (tmp == 255) ? 0xff : tmp - 1;
        } else {
          fprintf(stderr, "Error: WAN port number out of range 1..%u.\n", MAX_SWITCH_PORTS);
          exit(EXIT_FAILURE);
        }
        break;
      default:
        usage(argc, argv);
    }
  }
  if (optind >= argc)
    usage(argc, argv);

  initsockfd();

  /* info */
  if (strcmp(argv[optind], "info") == 0) {
    return do_info();
  /* get_switch_mode */
  } else if (strcmp(argv[optind], "get_switch_mode") == 0 ||
             strcmp(argv[optind], "gsm") == 0) {
    return do_get_switch_mode();
  /* set_switch_mode */
  } else if (strcmp(argv[optind], "set_switch_mode") == 0 ||
             strcmp(argv[optind], "ssm") == 0) {
    if (optind + 1 >= argc) {
      fprintf(stderr, "Missing required argument newmode.\n");
      usage(argc, argv);
    }
    for (c = 0; c < CPMAC_MODE_MAX_NO - 1; ++c)
      if (strcmp(cpmac_modes[c], argv[optind + 1]) == 0)
        break;
    if (c == CPMAC_MODE_MAX_NO - 1) {
      fprintf(stderr, "Invalid switch mode: %s\n", argv[optind + 1]);
      usage(argc, argv);
    }
    fprintf(stderr, "Setting switch_mode to '%s' (%u).\n", cpmac_modes[c], c + 1);
    param_ssm.cpmac_mode = c + 1;
    if (param_ssm.wan_vlan_number > 0) {
      unsigned int i;
      fprintf(stderr, "%u VLAN IDs passed through WAN:", param_ssm.wan_vlan_number);
      for (i = 0; i < param_ssm.wan_vlan_number; ++i)
        fprintf(stderr, " %u", param_ssm.wan_vlan_id[i]);
      fprintf(stderr, "\n");
    }
    return do_set_switch_mode(&param_ssm);
  /* set_switch_mode_special */
  } else if (strcmp(argv[optind], "set_switch_mode_special") == 0 ||
             strcmp(argv[optind], "ssms") == 0) {
    int next_is_devname = 1;
    for (c = optind + 1; c < argc &&
         param_ssms.number_of_devices < AVM_CPMAC_MAX_PORTS; ++c) {
      if (next_is_devname) {
        strncpy(param_ssms.device[param_ssms.number_of_devices].name,
                argv[c], AVM_CPMAC_MAX_DEVICE_NAME_LENGTH);
        next_is_devname = 0;
      } else {
        tmp = 0;
        if (sscanf(argv[c], "%x", &tmp) != 1) {
          fprintf(stderr, "Error: Can't parse port mask '%s'.\n", argv[c]);
          exit(EXIT_FAILURE);
        }
        param_ssms.device[param_ssms.number_of_devices].target_mask = tmp;
        ++param_ssms.number_of_devices;
        next_is_devname = 1;
      }
    }
    if ((! next_is_devname) || c == optind + 1) {
      fprintf(stderr, "Error: Wrong parameter count.\n");
      exit(EXIT_FAILURE);
    }
    return do_set_switch_mode_special(&param_ssms);
  /* get_switch_mode_current */
  } else if (strcmp(argv[optind], "get_switch_mode_current") == 0 ||
             strcmp(argv[optind], "gsmc") == 0) {
    return do_get_switch_mode_current();
  /* get_power_mode */
  } else if (strcmp(argv[optind], "get_power_mode") == 0 ||
             strcmp(argv[optind], "gpm") == 0) {
    return do_get_power_mode();
  /* set_power_mode */
  } else if (strcmp(argv[optind], "set_power_mode") == 0 ||
             strcmp(argv[optind], "spm") == 0) {
    unsigned int max_power_mode = sizeof(cpmac_adm_phy_power_modes) / sizeof(cpmac_adm_phy_power_modes[0]);

    if (optind + 1 >= argc) {
      fprintf(stderr, "Missing required argument PORTNUMBER.\n");
      usage(argc, argv);
    }
    if (sscanf(argv[optind + 1], "%u", &tmp) != 1) {
      fprintf(stderr, "Error: Can't parse port number '%s'.\n", argv[optind + 1]);
      exit(EXIT_FAILURE);
    }
    if (optind + 2 >= argc) {
      fprintf(stderr, "Missing required argument POWERMODE.\n");
      usage(argc, argv);
    }
    for (c = 0; c < max_power_mode; ++c)
      if (strcmp(cpmac_adm_phy_power_modes[c], argv[optind + 2]) == 0)
        break;
    if (c == max_power_mode) {
      fprintf(stderr, "Invalid power mode: %s\n", argv[optind + 2]);
      usage(argc, argv);
    }
    fprintf(stderr, "Setting port %u to power mode '%s'.\n", tmp, cpmac_adm_phy_power_modes[c]);
    return do_set_power_mode(tmp, c);
  /* get_port_media */
  } else if (strcmp(argv[optind], "get_port_media") == 0 ||
             strcmp(argv[optind], "gpme") == 0) {
    unsigned int port;

    if (optind + 1 < argc) {
      if (sscanf(argv[optind + 1], "%d", &c) != 1) {
        fprintf(stderr, "Error: Can't parse port number '%s'.\n", argv[optind + 1]);
        exit(EXIT_FAILURE);
      }
      return do_get_port_media(c);
    }
    if (count_ports(NULL, &tmp) != 0)
      exit(EXIT_FAILURE);
    for (port = 1; port <= tmp; ++port)
      c = c | do_get_port_media(port);
    if (c)
      exit(EXIT_FAILURE);
  /* set_port_media */
  } else if (strcmp(argv[optind], "set_port_media") == 0 ||
             strcmp(argv[optind], "spme") == 0) {
    int fc = FC_DONTTOUCH;

    if (optind + 1 >= argc) {
      fprintf(stderr, "Missing required argument PORTNUMBER.\n");
      usage(argc, argv);
    }
    if (sscanf(argv[optind + 1], "%u", &tmp) != 1) {
      fprintf(stderr, "Error: Can't parse port number '%s'.\n", argv[optind + 1]);
      exit(EXIT_FAILURE);
    }
    if (optind + 2 >= argc) {
      fprintf(stderr, "Missing required argument MEDIA.\n");
      usage(argc, argv);
    }
    for (c = 0; c < MEDIA_COUNT; ++c)
      if (strcmp(media_str[c], argv[optind + 2]) == 0)
        break;
    if (c == MEDIA_COUNT) {
      fprintf(stderr, "Invalid media type: %s\n", argv[optind + 2]);
      usage(argc, argv);
    }
    if (optind + 3 < argc) {
      for (fc = 0; fc < FC_MAX; ++fc)
        if (strcmp(flow_control_str[fc], argv[optind + 3]) == 0)
          break;
      if (fc == FC_MAX) {
        fprintf(stderr, "Invalid flow control: %s\n", argv[optind + 3]);
        usage(argc, argv);
      }
    }
    fprintf(stderr, "Setting media type for port %u to '%s'%s.\n",
            tmp, media_str[c], fc != FC_DONTTOUCH ? (fc == FC_ENABLE ? " (enable flow-control)" : " (disable flow-control)") : "");
    return do_set_port_media(tmp, c, fc);
  /* dump_reg */
  } else if (strcmp(argv[optind], "dump_reg") == 0) {
    unsigned int val;
    
    if (optind + 1 >= argc) {
      fprintf(stderr, "Missing required argument REGADDR.\n");
      usage(argc, argv);
    }
    if (sscanf(argv[optind + 1], "%u", &tmp) != 1) {
      fprintf(stderr, "Error: Can't parse register address '%s'.\n", argv[optind + 1]);
      exit(EXIT_FAILURE);
    }
    
    for (c = 0; c < 16; ++c) {
      switchreg(0, tmp + c, &val);
      printf("0x%04x: %08lx\n", tmp + c, val);
    }

  } else {
    fprintf(stderr, "Invalid command: %s", argv[optind]);
    usage(argc, argv);
  }
  exit(EXIT_SUCCESS);
}
